#define _POSIX_C_SOURCE 200809L

#include <common/common.h>
#include <lektor/config.h>
#include <lektor/database.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <dlfcn.h>
#include <stdio.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <ctype.h>

typedef int (*ini_handler)(volatile sqlite3 *db, const char *section, const char *name, const char *value);

static inline char *
strip(char *s)
{
    char *p = s + strlen(s);
    while (p > s && isspace(*(--p)))
        *p = 0;
    return s;
}

static inline char *
skip(char *s)
{
    while (*s && isspace(*s))
        s++;
    return s;
}

static inline int
ini_parse(const char *path, ini_handler handle, volatile sqlite3 *db)
{
    char *start, *end, *name, *value;
    char section[INI_MAX_SECTION_LEN], line[INI_MAX_LINE_LEN];
    int error = 0, linenum = 0, len;
    FILE *file = fopen(path, "r");
    if (!file) {
        LOG_ERROR_SCT("PARSER", "Failed to open config file '%s'", path);
        return 1;
    }

    /* Parse the file */
    while (NULL != fgets(line, INI_MAX_LINE_LEN, file)) {
        ++linenum;
        start = skip(strip(line));

        /* Skip comments */
        if (strspn(start, ";#"))
            continue;

        /* Handle sections */
        else if (start[0] == '[') {
            end = &start[1 + strcspn(start + 1, "]")];
            if (end[0] == ']') {
                end[0] = '\0';
                len = strlen(&start[1]);
                len = MIN(len + 1, INI_MAX_SECTION_LEN);
                memcpy(section, &start[1], len);
                section[INI_MAX_SECTION_LEN - 1] = '\0';
            }

            else {
                error = 1;
                LOG_ERROR_SCT("PARSER", "Invalid section name at line '%d'", linenum);
            }
        }

        /* Handle name[:=]name pair */
        else if (start[0]) {
            end = &start[1 + strcspn(start + 1, ":=")];
            if (end[0] == '=' || end[0] == ':') {
                end[0] = '\0';
                name   = strip(start);
                value  = &end[1];

                /* Find a comment */
                end = &value[strcspn(value, ";#")];
                if (end[0])
                    end[0] = '\0';

                /* Skip all spaces */
                value = skip(value);
                strip(value);

                /* Handle the SECTION, NAME[:=]VALUE */
                if (handle(db, section, name, value)) {
                    error = 1;
                    LOG_ERROR_SCT("PARSER", "Failed to '[handle] %s, %s{:,=}%s' at line '%d'",
                                  section, name, value, linenum);
                }
            }

            else {
                error = 1;
                LOG_ERROR_SCT("PARSER", "Invalid name[:=]value pair at line '%d'", linenum);
            }
        }
    }

    /* End of the function */
    fclose(file);
    if (error)
        LOG_ERROR_SCT("PARSER", "An error occured while parsing the file '%s'", path);
    return error;
}

int
load_so(const char *const mod_path, const char *const mod_init, void *mod)
{
    /* An init function takes two arguments: the module pointer and a void*
       which is the handler for the .so file. It the role of the module to
       store the handle pointer.
       Uppon successfull completion, the function shall return 0, and return
       a non zero value if something bad happened. */
    int (*module_set_function)(void *const, void *const);
    char *error;
    void *handle = dlopen(mod_path, RTLD_NOW);

    if (NULL == handle) {
        LOG_ERROR("libdl error in dlopen on file '%s': %s", mod_path, dlerror());
        return 1;
    }

    dlerror();
    *(void **) (&module_set_function) = dlsym(handle, mod_init);

    if ((error = dlerror()) != NULL) {
        LOG_ERROR("libdl error in dlsym on file '%s': %s\n", mod_path, error);
        return 1;
    }

    return ! module_set_function(mod, handle);
}

inline int
load_module_by_name(volatile sqlite3 *db, const char *name, void *mod)
{
    char mod_path[PATH_MAX], mod_load[INI_MAX_LINE_LEN];

    if (!database_config_get_text(db, name, "path", mod_path, PATH_MAX) ||
        !database_config_get_text(db, name, "load_function", mod_load, INI_MAX_LINE_LEN)) {
        LOG_ERROR("Module named %s is incomplete or is not defined in config file", name);
        return 1;
    }

    return load_so(mod_path, mod_load, mod);
}

inline static int
validate_conf(volatile sqlite3 *db)
{
#define CHK_OPTION(section, name)                                               \
    if (!database_config_exists(db, section, name)) {                           \
        LOG_ERROR("%s", "Missing option \""name"\" in section \""section"\"");  \
        return 1;                                                               \
    }

    CHK_OPTION("externals", "mkvpropedit");

    CHK_OPTION("server", "host");
    CHK_OPTION("server", "port");
    CHK_OPTION("server", "max_clients");

    CHK_OPTION("database", "kara_dir");
    CHK_OPTION("database", "db_path");

    CHK_OPTION("repo", "name");
    CHK_OPTION("repo", "url");
    CHK_OPTION("repo", "json");
    CHK_OPTION("repo", "id_json");
    CHK_OPTION("repo", "id_kara");

    CHK_OPTION("player", "autoclear");
    CHK_OPTION("player", "def_random");
    CHK_OPTION("player", "def_consume");
    CHK_OPTION("player", "def_single");
    CHK_OPTION("player", "def_repeat");
    CHK_OPTION("player", "module");
#undef CHK_OPTION

    return 0;
}

static int
handler(volatile sqlite3 *user, const char *section, const char *name, const char *value)
{
    RETURN_UNLESS(section && name && value, "I can't complete the database with incomplete lines", 1);
    RETURN_UNLESS(database_config_set(user, section, name, value), "Failed to update the database", 1);
    return 0;
}

int
config_detect_file(char *conf, size_t conf_len)
{
    bool is_malloc = false;
    struct passwd *pw = getpwuid(getuid());
    char *home;

    if (conf == NULL)
        return 1;

    memset(conf, 0, conf_len * sizeof(char));

    /* Try the current working dir config file. */
    if (getcwd(conf, conf_len - 1) != NULL) {
        strncat(conf, "/lektor.ini", conf_len - 1);
        LOG_INFO("Trying %s", conf);
        if (!access(conf, R_OK | F_OK))
            goto found;
    }

    /* Try the config file from the config directory. */
    home = getenv("XDG_CONFIG_HOME");
    if (home && strlen(home) < conf_len) {
        /* Skip the strncat to not append the '.config/' to
           the XDG_CONFIG_HOME which must already have this directory in it */
        strncat(conf, "/lektor/lektor.ini", conf_len - 1);
        goto skip_this_strcat;
    }
    if (!home || (strlen(home) >= conf_len))
        home = getenv("HOME");
    if (!home || (strlen(home) >= conf_len))
        home = pw->pw_dir;
    if (!home || (strlen(home) >= conf_len))
        goto no_config_directory;
    memcpy(conf, home, (strlen(home) + 1) * sizeof(char));
    strncat(conf, "/.config/lektor/lektor.ini", conf_len - 1);
skip_this_strcat:
    LOG_INFO("Trying %s", conf);
    if (!access(conf, R_OK | F_OK))
        goto found;

no_config_directory:
    /* Try the '/opt/lektor' file. */
    memcpy(conf, "/opt/lektor/lektor.ini", sizeof("/opt/lektor/lektor.ini"));
    LOG_INFO("Trying %s", conf);
    if (!access(conf, R_OK | F_OK))
        goto found;

    /* Try the '/usr/local/etc/lektor.ini' file. */
    memcpy(conf, "/usr/local/etc/lektor.ini", sizeof("/usr/local/etc/lektor.ini"));
    LOG_INFO("Trying %s", conf);
    if (!access(conf, R_OK | F_OK))
        goto found;

    /* Try the '/etc/lektor.ini' file. */
    memcpy(conf, "/etc/lektor.ini", sizeof("/etc/lektor.ini"));
    LOG_INFO("Trying %s", conf);
    if (!access(conf, R_OK | F_OK))
        goto found;

    /* Error */
    LOG_ERROR("An error occured with file %s: %s", conf, strerror(errno));
    if (is_malloc)
        free(conf);
    return 1;
found:
    LOG_INFO("Using file %s", conf);
    return 0;
}

int
config_new(volatile sqlite3 *db, const char *conf)
{
    if (ini_parse(conf, handler, db)) {
        LOG_ERROR("Failed to parse file %s", conf);
        return 1;
    }

    if (validate_conf(db)) {
        LOG_ERROR("Configuration file %s is incomplete", conf);
        return 1;
    }

    return 0;
}

int
config_open(volatile sqlite3 *db)
{
    char conf_file[PATH_MAX];
    int ret = 1;
    GOTO_IF(config_detect_file(conf_file, PATH_MAX), "Error while searching for a config file", error);
    GOTO_IF(config_new(db, conf_file), "Failed to read configuration file", error);
    ret = 0;
error:
    return ret;
}
