#define _POSIX_C_SOURCE 200809L

#include <common/common.h>
#include <lektor/database.h>

#include <strings.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>

bool
database_search_database_init(volatile sqlite3 *db, struct lkt_search *ret)
{
    RETURN_UNLESS(ret, "Exit because return pointer is NULL", false);
    static const char *SQL_STMT_TEMPLATE =
        "WITH content AS ("
        " SELECT kara.id AS id, string AS any_col, LENGTH(CAST(kara.id AS TEXT)) AS len"
        " FROM kara WHERE %s LIKE ?)"
        "SELECT id, any_col, (SELECT MAX(len) FROM content)"
        "FROM content LIMIT %d OFFSET %d;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    ret->type = lkt_search_database;

    /* Search part */
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TEMPLATE, ret->name, ret->msg_count, ret->continuation);
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->ka_rgx, error);
    ret->db = db;
    return true;
error:
    sqlite3_finalize(ret->stmt);
    return false;
}

bool
database_search_sticker_init(volatile sqlite3 *db, struct lkt_search *ret)
{
    /* No bound checks in strcats, should be fine. Possible SQL injection, depend on the `type`. */
    static const char *SQL_all_types =
        "SELECT name, sts.id, value FROM 'stickers' JOIN "
        "( SELECT id, sticker, value, 'kara' FROM 'stickers.kara'"
        "  UNION"
        "  SELECT id, sticker, value , 'plt' FROM 'stickers.plt'"
        ") AS sts"
        "ON sts.sticker = 'stickers'.id";
    static const char *SQL_one_type =
        "SELECT name, sts.id, value, '%s' "
        "FROM 'stickers.%s' AS sts "
        "JOIN 'stickers' "
        "ON 'stickers'.id = sts.sticker";
    char SQL[LKT_MAX_SQLITE_STATEMENT];
    ret->type = lkt_search_sticker;

    if (ret->st_type == NULL)
        memcpy(SQL, SQL_all_types, strlen(SQL_all_types) + 1);
    else
        sprintf(SQL, SQL_one_type, ret->st_type, ret->st_type);
    if (ret->st_uri != 0)
        sprintf(SQL + strlen(SQL), " AND sts.id = %d", ret->st_uri);
    if (ret->st_op != 0)
        sprintf(SQL + strlen(SQL), " AND sts.value %s %d",
                ret->st_op == '>' ? ">=" : ret->st_op == '<' ? "<=" : "=",
                ret->st_value);
    strcat(SQL, !ret->name ? ";" : " AND name = ?;");

    SQLITE_PREPARE(db, ret->stmt, SQL, error);
    if (ret->name)
        SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->name, error);
    return true;
error:
    sqlite3_finalize(ret->stmt);
    return false;
}

bool
database_search_queue_init(volatile sqlite3 *db, struct lkt_search *ret)
{
    RETURN_UNLESS(ret, "Exit because return pointer is NULL", false);
    static const char *SQL_STMT_TEMPLATE =
        "WITH content AS ("
        " SELECT kara_id AS id, string AS any_col, LENGTH(CAST(kara_id AS TEXT)) AS len"
        " FROM queue"
        " JOIN kara "
        " ON kara_id = kara.id AND %s LIKE ?)"
        "SELECT id, any_col, (SELECT MAX(len) FROM content)"
        "FROM content LIMIT %d OFFSET %d;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    ret->type = lkt_search_queue;

    /* Search part */
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TEMPLATE, ret->name, ret->msg_count, ret->continuation);
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->ka_rgx, error);
    ret->db = db;
    return true;
error:
    sqlite3_finalize(ret->stmt);
    return false;
}

bool
database_search_iter(struct lkt_search *item)
{
    const char *sql_row, *type;
    int id, code, id_len;
    RETURN_UNLESS(item, "Exit because item is NULL, end iterations", false);
    GOTO_UNLESS(item->call, "Call function is NULL, terminate the search", end);
    code = sqlite3_step(item->stmt);

    if (code == SQLITE_DONE)
        goto end;

    if (code != SQLITE_ROW) {
        LOG_ERROR_SCT("DB", "Step failed, expected a ROW or a DONE: %s", sqlite3_errmsg((sqlite3 *) item->db));
        goto end;
    }

    switch (item->type) {
    case lkt_search_database:
    case lkt_search_queue:
        id      = sqlite3_column_int(item->stmt, 0);
        sql_row = (const char *) sqlite3_column_text(item->stmt, 1);
        id_len  = sqlite3_column_int(item->stmt, 2);
        return ((lkt_search_database_func) item->call)(item->srv, item->c, id, id_len, sql_row);
    case lkt_search_sticker:
        sql_row = (const char *) sqlite3_column_text(item->stmt, 0);    /* Name  */
        id      = sqlite3_column_int(item->stmt, 1);                    /* Id    */
        code    = sqlite3_column_int(item->stmt, 2);                    /* Value */
        type    = (const char *) sqlite3_column_text(item->stmt, 3);    /* Type  */
        return ((lkt_search_sticker_func) item->call)(item->srv, item->c, sql_row, type, id, code);
    case lkt_search_playlist:
    default:
        LOG_WARN_SCT("DB", "Search operation %d is not implemented", item->type);
        goto end;
    }

end:
    sqlite3_finalize(item->stmt);
    return false;
}
