#define _POSIX_C_SOURCE 200809L
#define _DEFAULT_SOURCE

#include <common/common.h>

#include <lektor/cmd.h>
#include <lektor/config.h>
#include <lektor/mkv.h>
#include <lektor/database.h>
#include <lektor/net.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <sys/wait.h>

static noreturn inline void
fail(const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    exit(write(2, "\n", 1) == 0);
}

/* ----------------- *
 * The main function *
 * ----------------- */

static sqlite3 *db = NULL;
static char kara_dir[PATH_MAX], db_path[PATH_MAX], mkvpropedit[PATH_MAX], buf[100];

void
open_db(void)
{
    if (!database_new((volatile sqlite3 **) &db))
        fail("Failed to init memory db");

    if (config_open(db))
        fail("Could not load configuration");

    if (!database_config_get_text(db, "externals", "mkvpropedit", mkvpropedit, PATH_MAX))
        fail("Not found externals->mkvpropedit");

    if (!database_config_get_text(db, "database", "db_path", db_path, PATH_MAX))
        fail("Not found database->db_path");

    if (!database_config_get_text(db, "database", "kara_dir", kara_dir, PATH_MAX))
        fail("Not found database->kara_dir");
}

noreturn void
conf__(struct lkt_cmd_args *args)
{
    UNUSED(args);
    fprintf(stdout, "%s\n", lkt_default_config_file);
    LOG_INFO("%s", "You may redirect this output to ~/.config/lektor/lektor.ini");
    exit(EXIT_SUCCESS);
}

noreturn void
cat__(struct lkt_cmd_args *args)
{
    int i;
    struct kara_metadata data;
    for (i = 0; i < args->argc; ++i) {
        if (kara_metadata_read(&data, args->argv[i]))
            fail("Failed to read metadata of file %s", args->argv[i]);
        printf("Kara path: %s\n"
               "  Song source: %s\n"
               "  Song title: %s\n"
               "  Category: %s\n"
               "  Type: %s (number %d)\n"
               "  Author: %s\n"
               "  Language: %s\n",
               args->argv[i], data.source_name, data.song_name, data.category,
               data.song_type, data.song_number, data.author_name, data.language);
    }
    exit(EXIT_SUCCESS);
}

noreturn void
init_metadata__(struct lkt_cmd_args *args)
{
    UNUSED(args);
    open_db();
    metadata_set_directory(kara_dir, mkvpropedit);
    exit(EXIT_SUCCESS);
}

noreturn void
init_populate__(struct lkt_cmd_args *args)
{
    UNUSED(args);
    open_db();
    if (!database_open(db, db_path))
        fail("Failed to open database");
    database_update(db, kara_dir, false);
    sqlite3_close(db);
    exit(EXIT_SUCCESS);
}

noreturn void
init_file__(struct lkt_cmd_args *args)
{
    open_db();
    int i;
    for (i = 0; i < args->argc; ++i)
        metadata_set_file((char *) args->argv[i], mkvpropedit);
    exit(EXIT_SUCCESS);
}

noreturn void
init_file_prompt__(struct lkt_cmd_args *args)
{
    int i, num;
    struct kara_metadata data;
    for (i = 0; i < args->argc; ++i) {
        LOG_INFO("Asking for file '%s':", args->argv[i]);

        if (get_stdin_line(" song_name (TITLE): ", data.song_name, LEKTOR_TAG_MAX))
            fail("Input error");

        if (get_stdin_line(" source_name (NAME): ", data.source_name, LEKTOR_TAG_MAX))
            fail("Input error");

        if (get_stdin_line(" category (cdg, vo, ...): ", data.category, LEKTOR_TAG_MAX))
            fail("Input error");

        if (get_stdin_line(" type (OP, AMV, ...): ", data.song_type, LEKTOR_TAG_MAX))
            fail("Input error");

        if (get_stdin_line(" language (jp, fr, undefined, ...): ", data.language, LEKTOR_TAG_MAX))
            fail("Input error");

        if (get_stdin_line(" author_name (Your peudal): ", data.author_name, LEKTOR_TAG_MAX))
            fail("Input error");

        if (!get_stdin_line(" type's number(1, 2, ...): ", buf, 100))
            data.song_number = ((num = atoi(buf)) <= 0) ? 1 : num;
        else
            fail("Input error");

        kara_metadata_write(&data, args->argv[i], mkvpropedit);
        LOG_INFO("You may name this kara with the following name: '%s - %s%d - %s'",
                 data.source_name, data.song_type, data.song_number, data.song_name);
    }
    exit(EXIT_SUCCESS);
}

noreturn void
get__(struct lkt_cmd_args *args)
{
    struct kara_metadata data;
    struct lkt_repo repo;

    if (args->argc != 1)
        fail("Invalid argument");

    int i = atoi(args->argv[0]);
    open_db();
    if (!database_open(db, db_path))
        fail("Failed to open database");

    if (repo_new(&repo, db))
        fail("Cound not create repo");

    if (repo_get_id(&repo, i, &data))
        fail("Cound not download json for kara %d", i);

    printf("Kara id: %d\n"
           "  Song source: %s\n"
           "  Song title: %s\n"
           "  Category: %s\n"
           "  Type: %s (number %d)\n"
           "  Author: %s\n"
           "  Language: %s\n",
           i, data.source_name, data.song_name, data.category,
           data.song_type, data.song_number, data.author_name,
           data.language);
    exit(EXIT_SUCCESS);
}

noreturn void
init_database__(struct lkt_cmd_args *args)
{
    UNUSED(args);
    open_db();
    exit(!database_init(db_path));
}

noreturn void
download__(struct lkt_cmd_args *args)
{
    if (args->argc != 2)
        fail("Invalud argument");

    struct lkt_repo repo;
    struct kara_metadata data;
    int i = atoi(args->argv[0]);
    open_db();
    if (!database_open(db, db_path))
        fail("Failed to open database");

    if (repo_new(&repo, db))
        fail("Could not create the repo");

    if (repo_download_id_sync(&repo, i, args->argv[1], &data))
        fail("Cound not download json for kara %d", i);

    printf("Kara %d at %s\n"
           "  Song source: %s\n"
           "  Song title: %s\n"
           "  Category: %s\n"
           "  Type: %s (number %d)\n"
           "  Author: %s\n"
           "  Language: %s\n",
           i, args->argv[1], data.source_name, data.song_name, data.category,
           data.song_type, data.song_number, data.author_name, data.language);
    exit(EXIT_SUCCESS);
}

noreturn void
source__(struct lkt_cmd_args *args)
{
    if (args->argc != 0)
        fail("The source command takes no arguments");
    char value[INI_MAX_LINE_LEN];
    char player_mod[INI_MAX_LINE_LEN];
    open_db();
    if (!database_open(db, db_path))
        fail("Failed to open database");

#define CHK_OPTION(section, name)                                                   \
    if (!database_config_get_text(db, section, name, value, INI_MAX_LINE_LEN)) {    \
        LOG_ERROR("Failed to get value for option '[%s] %s'", section, name);       \
        exit(EXIT_FAILURE);                                                         \
    }                                                                               \
    printf("[%s]\t%s -> %s\n", section, name, value);

    CHK_OPTION("externals", "mkvpropedit");

    CHK_OPTION("server", "host");
    CHK_OPTION("server", "port");
    CHK_OPTION("server", "max_clients");

    CHK_OPTION("player", "module");
    memcpy(player_mod, value, INI_MAX_LINE_LEN);
    CHK_OPTION("player", "autoclear");
    CHK_OPTION("player", "def_random");
    CHK_OPTION("player", "def_consume");
    CHK_OPTION("player", "def_single");
    CHK_OPTION("player", "def_repeat");

    CHK_OPTION("repo", "name");
    CHK_OPTION("repo", "url");
    CHK_OPTION("repo", "json");
    CHK_OPTION("repo", "id_json");
    CHK_OPTION("repo", "id_kara");

    CHK_OPTION("database", "kara_dir");
    CHK_OPTION("database", "db_path");

    CHK_OPTION(player_mod, "path");
    CHK_OPTION(player_mod, "load_function");
#undef CHK_OPTION

    exit(EXIT_SUCCESS);
}

static struct lkt_cmd_opt options_init[] = {
    { .name = "database",   .call = init_database__  },
    { .name = "populate",   .call = init_populate__  },
    { .name = "metadata",   .call = init_metadata__  },
    { .name = "file",       .call = init_file__      },
    LKT_OPT_NULL,
};

noreturn void
init__(struct lkt_cmd_args *args)
{
    lkt_cmd_parse(options_init, args->argc, args->argv);
}

static struct lkt_cmd_opt options[] = {
    { .name = "init",       .call = init__     },
    { .name = "get",        .call = get__      },
    { .name = "download",   .call = download__ },
    { .name = "cat",        .call = cat__      },
    { .name = "conf",       .call = conf__     },
    { .name = "source",     .call = source__   },
    LKT_OPT_NULL,
};

int
main(int argc, const char **argv)
{
    executable_name = "lktadm";
    lkt_cmd_parse(options, argc - 1, argv + 1);
}
