#define _POSIX_C_SOURCE 200809L
#define _DEFAULT_SOURCE

#include <common/common.h>
#include <lektor/mkv.h>
#include <strings.h>
#include <stdbool.h>
#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <regex.h>

static const char *METADATA_TEMPLATE =
    "   <?xml version=\"1.0\" encoding=\"UTF-8\"?>  "
    "   <Tags>                                      "
    "   <Tag>                                       "
    "     <Targets>                                 "
    "       <TargetTypeValue>70</TargetTypeValue>   "
    "     </Targets>                                "
    "     <Simple>                                  "
    "       <Name>TITLE</Name>                      "
    "       <String>%s</String>                     "
    "     </Simple>                                 "
    "   </Tag>                                      "
    "   <Tag>                                       "
    "     <Simple>                                  "
    "       <Name>TITLE</Name>                      "
    "       <String>%s</String>                     "
    "     </Simple>                                 "
    "     <Simple>                                  "
    "       <Name>CONTENT_TYPE</Name>               "
    "       <String>%s</String>                     "
    "     </Simple>                                 "
    "     <Simple>                                  "
    "       <Name>ADDRESS</Name>                    "
    "       <String>%s</String>                     "
    "     </Simple>                                 "
    "     <Simple>                                  "
    "       <Name>ARTIST</Name>                     "
    "       <String>%s</String>                     "
    "     </Simple>                                 "
    "     <Simple>                                  "
    "       <Name>GENRE</Name>                      "
    "       <String>%s</String>                     "
    "     </Simple>                                 "
    "     <Simple>                                  "
    "       <Name>TRACK</Name>                      "
    "       <String>%d</String>                     "
    "     </Simple>                                 "
    "   </Tag>                                      "
    "   </Tags>                                     ";

static bool
mkvpropedit__(const char *const args[])
{
    pid_t pid;
    int wstatus, status, fd;

    if ((pid = fork()) == 0) {
        if ((fd = open("/dev/null", O_WRONLY | O_TRUNC)) < 0) {
            LOG_ERROR_SCT("FORK", "%s", "Can't to open /dev/null in O_WRONLY O_TRUNC");
            return false;
        }

        if (dup2(fd, 1) < 0) {
            LOG_ERROR_SCT("FORK", "%s", "Failed to duplicate /dev/null to stdout");
            return false;
        }

        execv(args[0], (char *const *) args);
        exit(EXIT_FAILURE);
    }

    else if (pid < 0) {
        LOG_ERROR_SCT("FORK", "Failed to fork: %s\n", strerror(errno));
        return false;
    }

    else {
        do {
            if (waitpid(pid, &wstatus, WUNTRACED | WCONTINUED) == -1) {
                LOG_ERROR_SCT("FORK", "Failed to wait children: %s", strerror(errno));
                return false;
            }
        } while (!WIFEXITED(wstatus) && !WIFSIGNALED(wstatus));

        if ((status = WEXITSTATUS(wstatus))) {
            LOG_ERROR_SCT("FORK", "Children failed with status %d", status);
            return false;
        }
    }

    return true;
}

int
kara_metadata_write(struct kara_metadata *mdt, const char *filename, const char *mkvpropedit)
{
    char tmpfilepath[PATH_MAX];
    char *const metadafilepath = &tmpfilepath[4];
    const char *const args1[] = { mkvpropedit, "-t", "all:", filename, NULL };
    const char *const args2[] = { mkvpropedit, "-t", tmpfilepath, filename, NULL };
    int fd, sta = -1;

    memset(tmpfilepath, 0, PATH_MAX);
    strncat(tmpfilepath, "all:/tmp/lektor.metadata.XXXXXX", PATH_MAX - 1);

    if ((fd = mkstemp(metadafilepath)) < 0) {
        LOG_ERROR("Failed to create temporary file: %s", strerror(errno));
        goto error;
    }

    if (dprintf(fd, METADATA_TEMPLATE, mdt->source_name, mdt->song_name, mdt->category,
                mdt->language, mdt->author_name, mdt->song_type, mdt->song_number) < 0) {
        LOG_ERROR("Failed to write to temporary file: %s", metadafilepath);
        goto error;
    }

    if (!mkvpropedit__(args1) || !mkvpropedit__(args2))
        goto error;

    sta = 0;
error:
    close(fd);
    unlink(metadafilepath);
    return sta;
}

static bool
metadata_from_path(char *const mkvfile, struct kara_metadata *meta)
{
    static regex_t regex;
    static int regex_init = 0;
    const size_t nmatch = 10;
    regmatch_t pmatch[nmatch];
    int reti, sta = false;
    char msgbuf[LEKTOR_TAG_MAX];
    char *endptr;

    static const char *rgx =
        "^/(.+)/(vo|va|amv|cdg|autres|vocaloid)/"
        "(jp|fr|en|ru|sp|it|ch|latin|multi|undefined)/(.+)/"
        "(.+) - (OP|ED|IS|AMV|VOCA|PV|MV|LIVE)([[:digit:]]*) - (.+)\\.mkv$";

    if (!regex_init)
        GOTO_IF(regcomp(&regex, rgx, REG_EXTENDED), "Failed to compile regex", error);

    memset(meta, 0, sizeof(struct kara_metadata));
    memset(msgbuf, 0, LEKTOR_TAG_MAX * sizeof(char));

    reti = regexec(&regex, mkvfile, nmatch, pmatch, 0);
    if (!reti) {
        /* Match */
        memcpy(meta->category, mkvfile + pmatch[2].rm_so, pmatch[2].rm_eo - pmatch[2].rm_so);
        memcpy(meta->language, mkvfile + pmatch[3].rm_so, pmatch[3].rm_eo - pmatch[3].rm_so);
        memcpy(meta->author_name, mkvfile + pmatch[4].rm_so, pmatch[4].rm_eo - pmatch[4].rm_so);
        memcpy(meta->source_name, mkvfile + pmatch[5].rm_so, pmatch[5].rm_eo - pmatch[5].rm_so);
        memcpy(meta->song_type, mkvfile + pmatch[6].rm_so, pmatch[6].rm_eo - pmatch[6].rm_so);
        memcpy(msgbuf, mkvfile + pmatch[7].rm_so, pmatch[7].rm_eo - pmatch[7].rm_so);
        memcpy(meta->song_name, mkvfile + pmatch[8].rm_so, pmatch[8].rm_eo - pmatch[8].rm_so);
    } else if (REG_NOMATCH == reti) {
        LOG_ERROR("No match for: %s", mkvfile);
        goto error;
    } else {
        regerror(reti, &regex, msgbuf, sizeof(msgbuf));
        LOG_ERROR("Failed to execute regex: %s", msgbuf);
        goto error;
    }

    STRTOL(meta->song_number, msgbuf, endptr, reti);
    if (reti || meta->song_number <= 0)
        meta->song_number = 1;

    sta = true;
error:
    /* regfree(&regex); // Should not be nescessary. */
    return sta;
}

int
metadata_set_directory(const char *kara_dir, const char *mkvpropedit)
{
    DIR *d;
    struct dirent *dir;
    char path[PATH_MAX];
    struct kara_metadata meta;
    memset(path, 0, PATH_MAX * sizeof(char));

    if (!(d = opendir(kara_dir))) {
        LOG_ERROR("Failed to open directory '%s': %s", kara_dir, strerror(errno));
        return 1;
    }

    while (NULL != (dir = readdir(d))) {
        strncpy(path, kara_dir, PATH_MAX - 1);
        strncat(path, "/", PATH_MAX - 1);
        strncat(path, dir->d_name, PATH_MAX - 1);

        if (dir->d_type == DT_REG               &&
            metadata_from_path(path, &meta)     &&
            ! kara_metadata_write(&meta, path, mkvpropedit))
            continue;
        else if (dir->d_type == DT_DIR          &&
                 !STR_MATCH(dir->d_name, ".")   &&
                 !STR_MATCH(dir->d_name, ".."))
            metadata_set_directory(path, mkvpropedit);
    }

    LOG_INFO("Passed directory '%s'", kara_dir);
    closedir(d);
    return false;
}

int
metadata_set_file(char *karapath, const char *mkvpropedit)
{
    struct kara_metadata meta;

    return metadata_from_path(karapath, &meta) &&
           ! kara_metadata_write(&meta, karapath, mkvpropedit);
}

