#define _POSIX_C_SOURCE 200809L

#include <mthread/mthread_internal.h>
#include <errno.h>
#include <stdlib.h>

/* TODO -> use binaty trees for better performance in searches. */
typedef struct cell_s {
    mthread_key_t key;
    const volatile void *volatile spec;
    void (* destr)(void *);
    volatile struct cell_s *volatile next;
} cell_t;

/* Default is 0. Index start at zero. Because of that the vairables hold the
   first available slots and not the last used slots in memory. */
static volatile mthread_key_t first_avail_key   = 1;
static volatile cell_t *volatile key_table      = NULL;

/* Protects anything concerning the keys. */
static mthread_tst_t lock = 0;

/* Functions for handling thread-specific data. */

/* Create a key value identifying a location in the thread-specific
   data area.  Each thread maintains a distinct thread-specific data
   area. DESTR_FUNCTION, if non-NULL, is called with the value
   associated to that key when the key is destroyed.
   DESTR_FUNCTION is not called if the value associated is NULL when
   the key is destroyed. */
int
mthread_key_create(mthread_key_t *__key, void (*__destr_function) (void *))
{
    cell_t *new_cell = (cell_t *) safe_malloc(sizeof(cell_t));
    mthread_spinlock_lock(&lock);

    /* Creation. */
    new_cell->key   = first_avail_key;
    new_cell->spec  = NULL;
    new_cell->destr = __destr_function;
    new_cell->next  = NULL;

    first_avail_key++;

    if (key_table) {
        LOG_INFO("Enqueue cell %d into global key list, old head is %d",
                 new_cell->key, key_table->key);
        new_cell->next = key_table;
        key_table = new_cell;
        LOG_INFO("New head is %d, next is %d",
                 key_table->key, key_table->next->key);
    } else {
        LOG_INFO("%s", "Set base cell for global key list");
        key_table = new_cell;
    }

    /* Register key into the thread. */

    struct keys_list *keys = &(mthread_self()->keys);
    if (!keys->list) {
        keys->list = (mthread_key_t *) safe_malloc(INIT_KEYS_LIST * sizeof(mthread_key_t));
        keys->first_avail = 0;
        keys->size = INIT_KEYS_LIST;
        LOG_INFO("%s", "Create the key list for thread");
    }

    if (keys->size == keys->first_avail) {
        /* Extend size if no space left. */
        void *new = realloc((void *) keys->list, (keys->size + INIT_KEYS_LIST) * sizeof(mthread_key_t));
        GOTO_UNLESS(new, "Could not increase the size of the key list", end);
        keys->size += INIT_KEYS_LIST;
        keys->list = new;
        LOG_INFO("%s", "Increase the size of the key list");
    }

    /* Register the key into the thread. */
    keys->list[keys->first_avail] = new_cell->key;
    keys->first_avail++;

    /* Returns. */
    *__key = new_cell->key;
    LOG_INFO("Created key %d", *__key);
end:
    mthread_spinlock_unlock(&lock);
    return errno;
}

/* Destroy KEY. */
int
mthread_key_delete(mthread_key_t __key)
{
    mthread_spinlock_lock(&lock);

    cell_t *it = (cell_t *) key_table;
    cell_t *last = NULL;

    while (it != NULL && it->key != __key) {
        last = it;
        it = (cell_t *) it->next;
    }

    if (!it) {
        errno = EINVAL;
        LOG_ERROR("Key %u not found", __key);
        goto end;
    }

    if (it->destr) {
        LOG_INFO("Call function on data at %p", (void *) it->spec);
        it->destr((void *) it->spec);
    }

    /* Handle deletion in the linked list. */

    if (last)
        last->next = it->next;
    else
        key_table = it->next;

    free(it);
    LOG_INFO("Deleted key %u successfully", __key);
    /* End of the function, simply returns. */
end:
    mthread_spinlock_unlock(&lock);
    return errno;
}

/* Store POINTER in the thread-specific data slot identified by KEY. */
int
mthread_setspecific(mthread_key_t __key, const void *__pointer)
{
    mthread_spinlock_lock(&lock);

    cell_t *it = (cell_t *) key_table;
    while (it != NULL && it->key != __key)
        it = (cell_t *) it->next;

    if (!it) {
        errno = EINVAL;
        LOG_ERROR("Key %u not found", __key);
        goto end;
    }

    it->spec = __pointer;
    LOG_INFO("Key %u set to %p", __key, __pointer);
    /* End of the function, simply returns. */
end:
    mthread_spinlock_unlock(&lock);
    return errno;
}

/* Return current value of the thread-specific data slot identified by KEY. */
void *
mthread_getspecific(mthread_key_t __key)
{
    void *ret = NULL;
    mthread_spinlock_lock(&lock);

    cell_t *it = (cell_t *) key_table;
    while (it != NULL && it->key != __key)
        it = (cell_t *) it->next;

    if (!it) {
        errno = EINVAL;
        LOG_ERROR("Key %u not found", __key);
        goto end;
    }

    ret = (void *) it->spec;
    /* End of the function, simply returns. */
end:
    mthread_spinlock_unlock(&lock);
    return (void *) ret;
}
