#define _POSIX_C_SOURCE 200809L

#include <common/common.h>
#include <lektor/thread.h>
#include <pthread.h>
#include <sys/types.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

struct __args {
    void *(*start)(struct poller_thread_arg *);
    struct poller_thread_arg *arg;
};

void *
__start(void *args__)
{
    struct __args *args = (struct __args *) args__;
    void *(*start)(struct poller_thread_arg *) = args->start;
    struct poller_thread_arg *arg = args->arg;
    free(args__);
    return start(arg);
}

int
poller_new(struct poller_thread *th, unsigned int init_sizes,
           void *(*func)(struct poller_thread_arg *), void *args)
{
    pthread_mutex_t mtx1 = PTHREAD_MUTEX_INITIALIZER;
    pthread_mutex_t mtx2 = PTHREAD_MUTEX_INITIALIZER;
    struct poller_thread th_ = {
        .initial_size   = init_sizes,
        .input_lock     = mtx1,
        .output_lock    = mtx2,
        .input_len      = 0,
        .output_len     = 0,
        .input_size     = init_sizes,
        .output_size    = init_sizes,
        .input_cells    = calloc(init_sizes, sizeof(void *)),
        .output_cells   = calloc(init_sizes, sizeof(void *)),
    };

    if (!th_.input_cells || !th_.output_cells)
        goto out_of_memory;

    *th = th_;
    struct __args *__args = calloc(1, sizeof(struct __args));
    __args->start = func;
    __args->arg = args;
    __args->arg->self = th;

    if (!pthread_create(&(th->th), NULL, __start, __args)) {
        LOG_INFO_SCT("THREAD", "%s", "Create a new poller thread");
        return 0;
    }

    else {
error:
        LOG_ERROR_SCT("THREAD", "%s", "Failed to create a poller thread");
        return 1;
    }

out_of_memory:
    LOG_ERROR_SCT("MEMORY", "%s", "Out of memory");
    if (th_.input_cells)
        free((void *) th_.input_cells);
    if (th_.output_cells)
        free((void *) th_.output_cells);
    goto error;
}

int
poller_join(struct poller_thread *th, void **ret)
{
    int sta = pthread_join(th->th, ret);

    if (sta)
        LOG_ERROR("%s", "Failed to join thread");

    if (th->input_cells)
        free((void *) th->input_cells);
    if (th->output_cells)
        free((void *) th->output_cells);

    if (!sta)
        LOG_ERROR("%s", "Thread joined");

    memset(th, 0, sizeof(struct poller_thread));
    return sta;
}

int
th_append(unsigned int *len, unsigned int *size, void ***cells,
          pthread_mutex_t *lock, void *ptr)
{
    void *new;
    int ret = 0;

    GOTO_IF(pthread_mutex_lock(lock), "Failed to lock", end);

    if (*len == *size) {
        new = realloc((void *) *cells, (*len + *size) * sizeof(void *));

        if (new == NULL) {
            LOG_ERROR_SCT("MEMORY", "%s", "Out of memory");
            ret = (errno = ENOMEM);
            goto end;
        }

        (*len) += *size;
        (*cells) = new;
    }

    (*cells)[(*len)++] = ptr;
    ret = 0;
end:
    RETURN_IF(pthread_mutex_unlock(lock), "Failed to lock", 1);
    return ret;
}

int
th_pop(unsigned int *len, void **cells, pthread_mutex_t *lock, void **ptr)
{
    int ret = 1;

    GOTO_IF(pthread_mutex_lock(lock), "Failed to lock", end);

    if (*len > 0)
        *ptr = cells[--(*len)];
    else
        *ptr = NULL;

    ret = 0;
end:
    RETURN_IF(pthread_mutex_unlock(lock), "Failed to lock", 1);
    return ret;
}

int
th_head(unsigned int len, void **cells, pthread_mutex_t *lock, void **ptr)
{
    int ret = 1;
    GOTO_IF(pthread_mutex_lock(lock), "Failed to lock", end);

    if (len > 0)
        *ptr = cells[len - 1];
    else
        *ptr = NULL;

    ret = 0;
end:
    RETURN_IF(pthread_mutex_unlock(lock), "Failed to lock", 1);
    return ret;
}

int
poller_append_input(struct poller_thread *th, void *ptr)
{
    return th_append((unsigned int *) &th->input_len, (unsigned int *) &th->input_size,
                     (void ***) &th->input_cells, &th->input_lock, ptr);
}

int
poller_append_output(struct poller_thread *th, void *ptr)
{
    return th_append((unsigned int *) &th->output_len, (unsigned int *) &th->output_size,
                     (void ***) &th->output_cells, &th->output_lock, ptr);
}

int
poller_pop_input(struct poller_thread *th, void **ptr)
{
    return th_pop((unsigned int *) &th->input_len, (void **) th->input_cells,
                  &th->input_lock, ptr);
}


int
poller_pop_output(struct poller_thread *th, void **ptr)
{
    return th_pop((unsigned int *) &th->output_len, (void **) th->output_cells,
                  &th->output_lock, ptr);
}

int
poller_head_input(struct poller_thread *th, void **ptr)
{
    return th_head((unsigned int) th->input_len, (void **) th->input_cells, &th->input_lock, ptr);
}

int
poller_head_output(struct poller_thread *th, void **ptr)
{
    return th_head((unsigned int) th->output_len, (void **) th->output_cells, &th->output_lock, ptr);
}
